package com.klipwallet.app2app.api;

import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.net.Uri;

import com.klipwallet.app2app.api.request.AuthRequest;
import com.klipwallet.app2app.api.request.CardTxRequest;
import com.klipwallet.app2app.api.request.ContractTxRequest;
import com.klipwallet.app2app.api.request.KlayTxRequest;
import com.klipwallet.app2app.api.request.KlipRequest;
import com.klipwallet.app2app.api.request.SignMessageRequest;
import com.klipwallet.app2app.api.request.TokenTxRequest;
import com.klipwallet.app2app.api.request.model.BAppInfo;
import com.klipwallet.app2app.api.response.CardListResponse;
import com.klipwallet.app2app.api.response.KlipResponse;
import com.klipwallet.app2app.common.KlipContextService;
import com.klipwallet.app2app.exception.KlipErrorCode;
import com.klipwallet.app2app.exception.KlipRequestException;
import com.klipwallet.app2app.internal.PrepareRequest;
import com.klipwallet.app2app.internal.Request;
import com.klipwallet.app2app.network.NetworkService;
import com.klipwallet.app2app.network.Utility;
import com.klipwallet.app2app.util.Logger;

import java.io.UnsupportedEncodingException;

/**
 * Klip API 제공하는 Class
 */
public class Klip {
    /**
     * Klip 인스턴스를 가져온다.
     *
     * @param context Application 컨텍스트
     * @return Klip 인스턴스
     */
    public synchronized static Klip getInstance(Context context) {
        if (instance == null) {
            instance = new Klip(context, NetworkService.Factory.getInstance());
            KlipContextService.getInstance().initialize(context);
        }
        return instance;
    }

    private Context context;
    private static Klip instance;
    private NetworkService networkService;

    Klip(Context context, NetworkService networkService) {
        this.context = context;
        this.networkService = networkService;
    }

    /**
     * Klip 요청 준비를 한다. (Request Key 획득)
     *
     * @param request  요청하는 실행 정보
     * @param bappInfo 요청하는 BApp 정보
     * @param callback 요청 결과를 받는 Callback (이때, T값은 KlipResponse을 사용)
     * @throws KlipRequestException 입력 데이터가 유효하지 않을 경우 발생
     */
    public void prepare(final KlipRequest request, final BAppInfo bappInfo, final KlipCallback callback) throws KlipRequestException {
        if (request == null || bappInfo == null || callback == null) {
            throw new KlipRequestException(KlipErrorCode.CORE_PARAMETER_MISSING, "required parameters is missing");
        }

        PrepareRequest req = new PrepareRequest();
        req.setBapp(bappInfo);

        if (request instanceof AuthRequest) {
            req.setType(KlipProtocol.AUTH);
        } else if (request instanceof KlayTxRequest) {
            req.setType(KlipProtocol.SEND_KLAY);
            req.setKlipRequest(request);
        } else if (request instanceof TokenTxRequest) {
            req.setType(KlipProtocol.SEND_TOKEN);
            req.setKlipRequest(request);
        } else if (request instanceof CardTxRequest) {
            req.setType(KlipProtocol.SEND_CARD);
            req.setKlipRequest(request);
        } else if (request instanceof ContractTxRequest) {
            req.setType(KlipProtocol.EXECUTE_CONTRACT);
            req.setKlipRequest(request);
        } else if (request instanceof SignMessageRequest) {
            req.setType(KlipProtocol.SIGN_MESSAGE);
            req.setKlipRequest(request);
        } else {
            throw new KlipRequestException(KlipErrorCode.NOT_SUPPORTED_REQUEST_TYPE, "not supported request type");
        }

        networkService.execute(getApiURL(KlipProtocol.API_PREPARE_URL), KlipProtocol.API_PREPARE_METHOD, req.toJson().toString(), KlipResponse.getConverter(), callback);
    }

    /**
     * Klip 요청 실행을 한다. (Deep Link 사용)
     *
     * @param requestKey 요청 키
     * @throws KlipRequestException 입력 데이터가 유효하지 않을 경우 발생
     */
    public void request(String requestKey) throws KlipRequestException {
        if (requestKey == null || requestKey.isEmpty()) {
            throw new KlipRequestException(KlipErrorCode.CORE_PARAMETER_MISSING, "requestKey is required");
        }
        Intent intent = new Intent(Intent.ACTION_VIEW);

        if (isKlipAppInstalled() && isAvailableKlipApp()) {
            intent.setData(Uri.parse("klip://klipwallet/open?url=" + getLinkURL() + requestKey));
        } else {
            Logger.info("klip is not available.");
            intent.setData(Uri.parse("https://klipwallet.com")); // 앱 다운로드 페이지
            intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);

        }
        context.startActivity(intent);
    }

    /**
     * Klip 요청에 대한 실행 결과 데이터를 가져온다.
     *
     * @param requestKey 요청 키
     * @param callback   요청 결과를 받는 Callback
     * @throws KlipRequestException 입력 데이터가 유효하지 않을 경우 발생
     */
    public void getResult(final String requestKey, final KlipCallback callback) throws KlipRequestException {
        if (requestKey == null || requestKey.length() == 0 || callback == null) {
            throw new KlipRequestException(KlipErrorCode.CORE_PARAMETER_MISSING, "required parameters is missing");
        }

        Request req = new Request();
        req.setRequestKey(requestKey);

        try {
            final String getParams = Utility.getQueryString(req.toMap(), "UTF-8");
            networkService.execute(getApiURL(KlipProtocol.API_RESULT_URL + "?" + getParams), KlipProtocol.API_RESULT_METHOD, null, KlipResponse.getConverter(), callback);
        } catch (UnsupportedEncodingException ignore) {
            // ignore
        }
    }

    /**
     * 사용자가 보유한 카드 정보 목록을 가져온다.
     *
     * @param cardAddress 조회하는 카드 주소
     * @param userAddress 조회하는 사용자 주소
     * @param cursor      조회하는 시작 커서
     * @param callback    요청 결과를 받는 Callback (이때, T값은 CardListResponse를 사용)
     * @throws KlipRequestException 입력 데이터가 유효하지 않을 경우 발생
     */
    public void getCardList(final String cardAddress, final String userAddress, final String cursor, final KlipCallback callback) throws KlipRequestException {
        if (cardAddress == null || cardAddress.length() == 0 ||
                userAddress == null || userAddress.length() == 0 || callback == null) {
            throw new KlipRequestException(KlipErrorCode.CORE_PARAMETER_MISSING, "required parameters is missing");
        }

        String getParams = ""
                + "sca=" + cardAddress
                + "&eoa=" + userAddress;
        if (cursor != null && cursor.length() > 0)
            getParams += "&cursor=" + cursor;
        networkService.execute(getApiURL(KlipProtocol.API_GET_CARD_URL + "?" + getParams), KlipProtocol.API_GET_CARD_METHOD, null, CardListResponse.getConverter(), callback);
    }

    /**
     * KlipApp 설치 여부를 확인한다.
     *
     * @return KlipApp 설치 여부
     */
    public boolean isKlipAppInstalled() {
        return isPackageInstalled(KlipProtocol.KLIP_APP_PACKAGE);
    }

    /**
     * 설치된 KlipApp과 연동 가능한지 버전을 확인한다.
     *
     * @return KlipApp 사용가능 여부
     */
    public boolean isAvailableKlipApp() {
        return isSupportedKlipWallet(KlipProtocol.KLIP_APP_PACKAGE, KlipProtocol.KLIP_APP_SUPPORT_VERSION);
    }

    private String getApiURL(String path) {
        return KlipProtocol.SCHEME + "://" + KlipProtocol.apiAuthority() + KlipProtocol.API_VERSION + path;
    }

    private String getLinkURL() {
        return KlipProtocol.SCHEME + "://" + KlipProtocol.linkAuthority() + KlipProtocol.KAKAO_KLIP_LINK;
    }

    private boolean isPackageInstalled(final String packageName) {
        return context.getPackageManager().getLaunchIntentForPackage(packageName) != null;
    }

    private boolean isSupportedKlipWallet(final String packageName, final int minVersion) {
        PackageInfo talkPackageInfo;
        try {
            talkPackageInfo = context.getPackageManager().getPackageInfo(packageName,
                    PackageManager.GET_SIGNATURES);
            return talkPackageInfo.versionCode >= minVersion;
        } catch (PackageManager.NameNotFoundException e) {
            return false;
        }
    }
}
